#include <Arduino.h>
// #define DTR_DEBUG_
#include "TinyDecisionTreeClassifier.h"
#include "ProcessingFunctions.h"

typedef float Treetype;
#include "Breast_cancer.h"

#define MAX_TREE_DEPTH        4
#define MIN_SAMPLES_SPLIT     10
#define TEST_RATIO            0.33f
#define TEST_SIZE             (uint32_t)(NUMBER_OF_SAMPLES*TEST_RATIO)
#define TRAIN_SIZE            (uint32_t)(NUMBER_OF_SAMPLES-TEST_SIZE)

// Treetype X_const[NUMBER_OF_SAMPLES][NUMBER_OF_FEATURES];
Treetype Y_const[NUMBER_OF_SAMPLES][1];

Treetype **X;
Treetype **Y;

#define BENCHMARK_AVERAGING   20
void benchmark(void){
  uint32_t timeBefore;
  uint32_t timeAfter;
  uint64_t benchmarkingTrainingTime = 0;
  uint64_t benchmarkingPredictionTime = 0;
  float totalScore=0;
  Serial.println("Bechmarking start");
  for(uint32_t i=0;i<BENCHMARK_AVERAGING;i++){
    shuffleData<Treetype>(X,Y,NUMBER_OF_SAMPLES,NUMBER_OF_FEATURES);
    TinyDecisionTreeClassifier<Treetype> clf(MAX_TREE_DEPTH,MIN_SAMPLES_SPLIT,TinyDecisionTreeClassifier<Treetype> ::GINI);
    timeBefore = micros();
    clf.fit(X+TEST_SIZE,Y+TEST_SIZE,TRAIN_SIZE,NUMBER_OF_FEATURES);
    // Serial.println("Tree before pruning: ");
    // clf.plot();
    while(clf.prune(X,Y,TRAIN_SIZE));
    timeAfter = micros();
    benchmarkingTrainingTime+=(timeAfter-timeBefore);
    Treetype rslt;
    for(uint32_t j=0;j<TRAIN_SIZE;j++){
        timeBefore = micros();
        rslt = clf.predict(X[j]);
        timeAfter = micros();
        benchmarkingPredictionTime=(timeAfter-timeBefore);
    }
    clf.plot();
    totalScore+=clf.score(X,Y,TEST_SIZE);
  }
  Serial.print("Score:");
  Serial.println(totalScore/BENCHMARK_AVERAGING);
  Serial.print("Training time: ");
  Serial.print((uint32_t)benchmarkingTrainingTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("Prediction time: ");
  Serial.print((uint32_t)benchmarkingPredictionTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("After averaging ");
  Serial.print(BENCHMARK_AVERAGING);
  Serial.println(" sessions.");
}

void setup() {
  Serial.begin(115200);
  delay(1000);
  X = new Treetype *[NUMBER_OF_SAMPLES];
  Y = new Treetype *[NUMBER_OF_SAMPLES];
  for(uint32_t i=0;i<NUMBER_OF_SAMPLES;i++){
      X[i] = X_const[i];
      Y[i] = Y_const[i];
      Y_const[i][0]=y[i];
  }
  benchmark();
}

void loop() {
  // Serial.println("Begin!");
  // digitalWrite(13,HIGH);
  // delay(100);
  // digitalWrite(13,LOW);
  delay(100);
}
