#include <Arduino.h>
// #define DTR_DEBUG_
#include "TinyDecisionTreeClassifier.h"

typedef float Treetype;
#include "Iris_flowers.h"

#define MAX_TREE_DEPTH      4
#define MIN_SAMPLES_SPLIT   10

// Treetype X_const[DATASET_SIZE][FEATURES_DIM];
Treetype Y_const[DATASET_SIZE][1];

Treetype **X;
Treetype **Y;

#define BENCHMARK_AVERAGING   10
void benchmark(void){
  uint32_t timeBefore;
  uint32_t timeAfter;
  uint64_t benchmarkingTrainingTime = 0;
  uint64_t benchmarkingPredictionTime = 0;
  Serial.println("Bechmarking start");
  for(uint32_t i=0;i<BENCHMARK_AVERAGING;i++){
    TinyDecisionTreeClassifier<Treetype> clf(MAX_TREE_DEPTH,MIN_SAMPLES_SPLIT,TinyDecisionTreeClassifier<Treetype>::INFO_GAIN);
    timeBefore = micros();
    clf.fit(X,Y,DATASET_SIZE,FEATURES_DIM);
    Serial.println("Tree before pruning: ");
    clf.plot();
    while(clf.prune(X,Y,DATASET_SIZE));
    timeAfter = micros();
    benchmarkingTrainingTime+=(timeAfter-timeBefore);
    Treetype rslt;
    for(uint32_t j=0;j<DATASET_SIZE;j++){
        timeBefore = micros();
        rslt = clf.predict(X[j]);
        timeAfter = micros();
        benchmarkingPredictionTime=(timeAfter-timeBefore);
    }
    clf.plot();
    Serial.print("Score:");
    Serial.println(clf.score(X,Y,DATASET_SIZE));
  }
  Serial.print("Training time: ");
  Serial.print((uint32_t)benchmarkingTrainingTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("Prediction time: ");
  Serial.print((uint32_t)benchmarkingPredictionTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("After averaging ");
  Serial.print(BENCHMARK_AVERAGING);
  Serial.println(" sessions.");
}

void setup() {
  Serial.begin(115200);
  delay(1000);
  X = new Treetype *[DATASET_SIZE];
  Y = new Treetype *[DATASET_SIZE];
  for(uint32_t i=0;i<DATASET_SIZE;i++){
      X[i] = input_data[i];
      Y[i] = Y_const[i];
      Y_const[i][0]=target_data[i];
  }
  benchmark();
}

void loop() {
  delay(100);
}
