#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BNO055.h>
#include <utility/imumaths.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <AceButton.h>
// #define DTR_DEBUG_
#include "TinyDecisionTreeClassifier.h"
#include "ProcessingFunctions.h"
#include "globals.h"

using namespace ace_button;
#define LED_PIN 8
#include "sdHelpFunctions.h"
#include "computationFunctions.h"

Adafruit_BNO055 bno = Adafruit_BNO055(55, 0x28, &Wire);

enum SystemState{
  IDLE=0x00,
  RECORDING_1=0x01,
  RECORDING_2=0x02,
  LEARNING=0x03,
  PREDICTING=0x04
}systemState;

enum ButtonEv{
  B_IDLE=0x00,
  B_CLICKED=0x01,
  B_DOUBLE_CLICKED=0x02,
  B_LONG_PRESSED=0x03
}buttonEv;

AceButton button;
void handleEvent(AceButton*, uint8_t, uint8_t);

void setup() {
  Serial.begin(115200);
  pinMode(22,OUTPUT);
  digitalWrite(22,HIGH);
  pinMode(LED_PIN,OUTPUT);
  digitalWrite(LED_PIN,HIGH);
  delay(1000);

  Wire.begin(21,20);

  SPI.begin(6,2,7,5);
  // SPI.setFrequency(10000);
  if(!SD.begin(5)){
    Serial.println("Card Mount Failed");
    return;
  }else{
    Serial.println("SD ok");
  }
  if(!bno.begin())
    {
      Serial.print("Ooops, no BNO055 detected ... Check your wiring or I2C ADDR!");
      // while(1);
    }

  delay(1000);
  bno.setExtCrystalUse(true);
  pinMode(GPIO_NUM_9, INPUT);
  button.init(GPIO_NUM_9, HIGH);
  ButtonConfig* buttonConfig = button.getButtonConfig();
  buttonConfig->setEventHandler(handleEvent);
  buttonConfig->setFeature(ButtonConfig::kFeatureClick);
  buttonConfig->setFeature(ButtonConfig::kFeatureDoubleClick);
  buttonConfig->setFeature(ButtonConfig::kFeatureLongPress);
  buttonConfig->setFeature(ButtonConfig::kFeatureRepeatPress);
  buttonConfig->setDoubleClickDelay(2000);
  // buttonConfig->setPressedDelay(10000);

  systemState = IDLE;
  buttonEv = B_IDLE;
  }

uint32_t windowCounter=0;
Treetype Xsample[9];
bool treeChecked = false;
void loop() {
  uint32_t i=0;
  switch (systemState)
  {
  case IDLE:
    digitalWrite(LED_PIN,HIGH);
    if(!treeChecked){
      if(fileExists(SD,"/tree.csv")){
        loadTree(SD,"/tree.csv");
        systemState=PREDICTING;
        allocBuffer(&X_window,WINDOW_SIZE,TREE_NUMBER_OF_RECORDED_FEATURES);
        allocBuffer(&Y_window,WINDOW_SIZE,1);
      }
      treeChecked=true;
    }
    if(buttonEv==B_CLICKED){
      for(i=0;i<100;i++){
        delay(10);
        button.check();
      }
      switch (buttonEv)
      {
      case B_CLICKED:
        systemState=RECORDING_1;
        buttonEv=B_IDLE;
        break;
      case B_DOUBLE_CLICKED:
        systemState=RECORDING_2;
        buttonEv=B_IDLE;
        break;
      case B_LONG_PRESSED:
        systemState=LEARNING;
        buttonEv=B_IDLE;
        break;   
      }
    }
    if(buttonEv==B_DOUBLE_CLICKED){
      systemState=RECORDING_2;
      buttonEv=B_IDLE;
      break;
    }
    if(buttonEv==B_LONG_PRESSED){
      systemState=LEARNING;
      buttonEv=B_IDLE;
      break;
    }
    break;
  case RECORDING_1:{
    allocBuffer(&X_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES,TREE_NUMBER_OF_RECORDED_FEATURES);
    allocBuffer(&Y_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES,1);

    Serial.println("Record class 1");
    for(i=0;i<TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES;i++){
      sensors_event_t event;
      bno.getEvent(&event,Adafruit_BNO055::VECTOR_ACCELEROMETER);
      X_train[i][0]=event.acceleration.x;
      X_train[i][1]=event.acceleration.y;
      X_train[i][2]=event.acceleration.z;
      Y_train[i][0]=1;
      delay(8);
      if(i%10==0)digitalWrite(LED_PIN,!digitalRead(LED_PIN));
    }
    saveNewDataForClass1();
    systemState=IDLE;
    deallocBuffer(&X_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES);
    deallocBuffer(&Y_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES);
  break;
  }
  case RECORDING_2:{
    Serial.println("Record class 2");
    allocBuffer(&X_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES,TREE_NUMBER_OF_RECORDED_FEATURES);
    allocBuffer(&Y_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES,1);
    for(i=0;i<TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES;i++){
      sensors_event_t event;
      bno.getEvent(&event,Adafruit_BNO055::VECTOR_ACCELEROMETER);
      X_train[i][0]=event.acceleration.x;
      X_train[i][1]=event.acceleration.y;
      X_train[i][2]=event.acceleration.z;
      Y_train[i][0]=0;
      delay(8);
      if(i%100==0)digitalWrite(LED_PIN,!digitalRead(LED_PIN));
    }
    saveNewDataForClass2();
    systemState=IDLE;
    deallocBuffer(&X_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES);
    deallocBuffer(&Y_train,TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES);
  break;
  }
  case LEARNING:{
    int32_t rslt = train();
    //dealoc buffers
    systemState=PREDICTING;
    allocBuffer(&X_window,WINDOW_SIZE,TREE_NUMBER_OF_RECORDED_FEATURES);
    allocBuffer(&Y_window,WINDOW_SIZE,1);

  break;
  }
  case PREDICTING:{
    digitalWrite(LED_PIN,LOW);
    sensors_event_t event;
    bno.getEvent(&event,Adafruit_BNO055::VECTOR_ACCELEROMETER);
    // Serial.println("Predicting");
    X_window[windowCounter][0]=event.acceleration.x;
    X_window[windowCounter][1]=event.acceleration.x;
    X_window[windowCounter][2]=event.acceleration.x;
    windowCounter++;
    if(windowCounter==WINDOW_SIZE){
      computeFeatures(X_window,&Xsample[0],&Xsample[1],&Xsample[2],0,WINDOW_SIZE);
      computeFeatures(X_window,&Xsample[3],&Xsample[4],&Xsample[5],1,WINDOW_SIZE);
      computeFeatures(X_window,&Xsample[6],&Xsample[7],&Xsample[8],2,WINDOW_SIZE);
      if(clf.root!=NULL){
        Serial.println(clf.predict(Xsample));
      }
      windowCounter=0;
    }

    if(buttonEv==B_CLICKED){
      for(i=0;i<150;i++){
        delay(10);
        button.check();
      }
      switch (buttonEv)
      {
      case B_LONG_PRESSED:
        deallocBuffer(&X_window,WINDOW_SIZE);
        deallocBuffer(&Y_window,WINDOW_SIZE);
        systemState=IDLE;
        buttonEv=B_IDLE;
        break;   
      default:
      {
        buttonEv=B_IDLE;
      }
      }
    }

    // systemState=PREDICTING;
  break;
  }
  default:
    break;
  }
  button.check();
  delay(5);
}


void handleEvent(AceButton* /* button */, uint8_t eventType,
    uint8_t buttonState) {

  Serial.print(F("handleEvent(): eventType: "));
  Serial.print(AceButton::eventName(eventType));
  Serial.print(F("; buttonState: "));
  Serial.println(buttonState);

  switch (eventType) {
    case AceButton::kEventPressed:
      buttonEv=B_CLICKED;
      Serial.println("Pressed");
      break;
    case AceButton::kEventClicked:
      buttonEv=B_CLICKED;
      Serial.println("Clicked");
      // digitalWrite(LED_PIN, LED_ON);
      break;
    case AceButton::kEventDoubleClicked:
      buttonEv=B_DOUBLE_CLICKED;
      Serial.println("Double Clicked");
      // digitalWrite(LED_PIN, LED_OFF);
      break;
    case AceButton::kEventLongPressed:
      buttonEv=B_LONG_PRESSED;
      Serial.println("Long Pressed");
      // digitalWrite(LED_PIN, LED_OFF);
      break;
  }
}