#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include "globals.h"
#include "computationFunctions.h"

bool fileExists(fs::FS &fs, const char * path){
   Serial.printf("Checking file file: %s\n", path);

   File file = fs.open(path);
   if(!file){
       return false;
   }
   file.close();
   return true;
}

void createDir(fs::FS &fs, const char * path){
    Serial.printf("Creating Dir: %s\n", path);
    if(fs.mkdir(path)){
        Serial.println("Dir created");
    } else {
        Serial.println("mkdir failed");
    }
}

void writeDataToFile(fs::FS &fs, const char * path){
    Serial.printf("Writing file: %s\n", path);

    File file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }
    uint32_t i=0;
    for(i=0;i<TREE_NUMBER_OF_RECORDED_TRAINING_SAMPLES;i++){
      // delay(2);
      if(file.printf("%f,%f,%f,%f,\n",X_train[i][0],X_train[i][1],X_train[i][2],Y_train[i][0])){
      } else {
        Serial.println("Write failed");
        Serial.print(X_train[i][0]);
        Serial.print(",");
        Serial.print(X_train[i][1]);
        Serial.print(",");
        Serial.print(X_train[i][2]);
        Serial.print(",");
        Serial.println(Y_train[i][0]);
        Serial.print("Write failure for raw:");
        Serial.println(i);
      }
      if(i%5==0)digitalWrite(LED_PIN,!digitalRead(LED_PIN));
  
    }
    file.close();
    Serial.println("Write success");
}

void saveNewDataForClass1(void){
    createDir(SD,"/class_1");
    uint32_t fileCtr=0;
    String fileName=String("/class_1/num_")+String(fileCtr)+String("_class_1.csv");
    while(fileExists(SD,fileName.c_str())){
      fileCtr++;
      fileName=String("/class_1/num_")+String(fileCtr)+String("_class_1.csv");
    }
    writeDataToFile(SD,fileName.c_str());
    }

void saveNewDataForClass2(void){
    createDir(SD,"/class_2");
    uint32_t fileCtr=0;
    String fileName=String("/class_2/num_")+String(fileCtr)+String("_class_2.csv");
    while(fileExists(SD,fileName.c_str())){
      fileCtr++;
      fileName=String("/class_2/num_")+String(fileCtr)+String("_class_2.csv");
    }
    writeDataToFile(SD,fileName.c_str());
    }


uint32_t countSamples(fs::FS &fs, const char * dirname){
    Serial.printf("Counting directory: %s\n", dirname);
    uint32_t lineCount = 0;

    File root = fs.open(dirname);
    if(!root){
        Serial.println("Failed to open directory");
        return 0;
    }
    if(!root.isDirectory()){
        Serial.println("Not a directory");
        return 0;
    }

    File file = root.openNextFile();
    while(file){
        digitalWrite(LED_PIN,!digitalRead(LED_PIN));
        if(file.isDirectory()){
            // Serial.print("  DIR : ");
            // Serial.println(file.name());
            // if(levels){
            //     listDir(fs, file.path(), levels -1);
            // }
        } else {
            Serial.print("  FILE: ");
            Serial.println(file.name());
            while (file.available()) {
              file.readStringUntil('\n');
              lineCount++;
            }
        }
        file = root.openNextFile();
    }
    return lineCount;
}

int32_t readTrainingData(fs::FS &fs, const char * dirname, uint32_t maxNumberOfSamples,uint32_t class_num,uint32_t readFrom){
    Serial.printf("Processing directory: %s\n", dirname);
    uint32_t samples_filled = readFrom;

    File root = fs.open(dirname);
    if(!root){
        Serial.println("Failed to open directory");
        return 0;
    }
    if(!root.isDirectory()){
        Serial.println("Not a directory");
        return 0;
    }

    File file = root.openNextFile();
    while(file){
        if(file.isDirectory()){
            // Serial.print("  DIR : ");
            // Serial.println(file.name());
            // if(levels){
            //     listDir(fs, file.path(), levels -1);
            // }
        } else {
            Serial.print("  FILE: ");
            Serial.println(file.name());
            uint32_t windowSampleCounter=0;
            while (file.available()) {
              String line = file.readStringUntil('\n');
            //   Serial.println(line.c_str());
            //   Serial.println(sscanf(line.c_str(), "%f,%f,%f,%f,\n", &(X_window[windowSampleCounter][0]),&(X_window[windowSampleCounter][1]),&(X_window[windowSampleCounter][2]),Y_window[windowSampleCounter][0]));
              if(sscanf(line.c_str(), "%f,%f,%f,%f,\n", &X_window[windowSampleCounter][0],&X_window[windowSampleCounter][1],&X_window[windowSampleCounter][2],&Y_window[windowSampleCounter][0])==4){
                // Serial.println("Scanf matched");
                windowSampleCounter++;
                if(windowSampleCounter==WINDOW_SIZE){
                  digitalWrite(LED_PIN,!digitalRead(LED_PIN));
                  computeFeatures(X_window,&X_train[samples_filled][0],&X_train[samples_filled][1],&X_train[samples_filled][2],0,WINDOW_SIZE);
                  computeFeatures(X_window,&X_train[samples_filled][3],&X_train[samples_filled][4],&X_train[samples_filled][5],1,WINDOW_SIZE);
                  computeFeatures(X_window,&X_train[samples_filled][6],&X_train[samples_filled][7],&X_train[samples_filled][8],2,WINDOW_SIZE);
                  Y_train[samples_filled][0]=class_num;
                  Serial.print("Window ");
                  Serial.print(samples_filled);
                  Serial.print(" ready:");
                  for(uint32_t k=0;k<TRAINING_FEATURES_SIZE;k++){
                      Serial.print(X_train[samples_filled][k],6);
                      Serial.print(",");
                  }
                  Serial.println(Y_train[samples_filled][0]);
                  windowSampleCounter=0;
                  samples_filled++;
                }
                if(samples_filled==maxNumberOfSamples){
                    file.close();
                    return samples_filled;
                }
              }
            }
        }
        file = root.openNextFile();
    }
    return samples_filled;
}

void saveTree(fs::FS &fs, const char * path){
    uint32_t outputdatasize = 100;
    char* outputdata = (char*) malloc(outputdatasize);
    clf.save(&outputdata,&outputdatasize);

    File file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }
    uint32_t i=0;
    Serial.println(outputdata);
    Serial.println(file.printf("%s",outputdata));
    file.close();
    Serial.println("Write success");
    free(outputdata);
}

void loadTree(fs::FS &fs, const char * path){
    Serial.printf("Reading tree from file: %s\n", path);

    File file = fs.open(path);
    if(!file){
        Serial.println("Failed to open file for reading");
        return;
    }

    Serial.print("Read from file: ");
    String tree=file.readString();
    Serial.println(tree.c_str());
    file.close();
    clf.load((char*)tree.c_str());
    clf.plot();
    Serial.println("Tree loaded");
}

int32_t train(void){
    uint32_t class1SamplesSize = countSamples(SD,"/class_1");
    uint32_t class2SamplesSize = countSamples(SD,"/class_2");
    uint32_t minimumSampleSize = class2SamplesSize>class1SamplesSize ? class1SamplesSize : class2SamplesSize;//doublecheck

    uint32_t numberOfWindowsNeeded = (minimumSampleSize/WINDOW_SIZE)*2;

    Serial.println(class1SamplesSize);
    Serial.println(class2SamplesSize);
    Serial.println(numberOfWindowsNeeded);

    if(numberOfWindowsNeeded<MAXIMUM_TRAINING_BUFFER_SIZE){
        allocBuffer(&X_train,numberOfWindowsNeeded,TRAINING_FEATURES_SIZE);
        allocBuffer(&Y_train,numberOfWindowsNeeded,1);
    }else{
        allocBuffer(&X_train,MAXIMUM_TRAINING_BUFFER_SIZE,TRAINING_FEATURES_SIZE);
        allocBuffer(&Y_train,MAXIMUM_TRAINING_BUFFER_SIZE,1);
    }
    allocBuffer(&X_window,WINDOW_SIZE,TREE_NUMBER_OF_RECORDED_FEATURES);
    allocBuffer(&Y_window,WINDOW_SIZE,1);

    uint32_t samplesRead = readTrainingData(SD,"/class_1",numberOfWindowsNeeded/2,1,0);
    samplesRead = readTrainingData(SD,"/class_2",numberOfWindowsNeeded,2,samplesRead);

    uint32_t testSize = samplesRead*TEST_RATIO;

    Serial.println("Preprocess finished");
    if(samplesRead>0){
        shuffleData<Treetype>(X_train,Y_train,samplesRead,TRAINING_FEATURES_SIZE);
        Serial.println("Fitting start");
        uint32_t fitting_start=micros();
        clf.fit(X_train+testSize,Y_train+testSize,samplesRead-testSize,TRAINING_FEATURES_SIZE);
        clf.plot();
        Serial.println(micros()-fitting_start);
        while(clf.prune(X_train+testSize,Y_train+testSize,samplesRead-testSize));      
        clf.plot();
        saveTree(SD,"/tree.csv");
        Serial.print("Score:");
        Serial.println(clf.score(X_train,Y_train,testSize));
    }

    deallocBuffer(&X_train,numberOfWindowsNeeded);
    deallocBuffer(&Y_train,numberOfWindowsNeeded);
    deallocBuffer(&X_window,WINDOW_SIZE);
    deallocBuffer(&Y_window,WINDOW_SIZE);

    return samplesRead;
}

