#include <Arduino.h>
// #define DTR_DEBUG_
#include "TinyDecisionTreeClassifier.h"
#include "FS.h"
#include "SD.h"
#include "SPI.h"

#define MAX_TREE_DEPTH      4
#define MIN_SAMPLES_SPLIT   10

#define NUMBER_OF_FEATURES  4
#define NUMBER_OF_TRAIN_SAMPLES   105
#define NUMBER_OF_TEST_SAMPLES    45

typedef float Treetype;
Treetype X_train_dat[NUMBER_OF_TRAIN_SAMPLES][NUMBER_OF_FEATURES];
Treetype Y_train_dat[NUMBER_OF_TRAIN_SAMPLES][1];
Treetype **X_train_ptr;
Treetype **Y_train_ptr;

Treetype X_test_dat[NUMBER_OF_TEST_SAMPLES][NUMBER_OF_FEATURES];
Treetype Y_test_dat[NUMBER_OF_TEST_SAMPLES][1];
Treetype **X_test_ptr;
Treetype **Y_test_ptr;

#define BENCHMARK_AVERAGING   10
void benchmark(void){
  uint32_t timeBefore;
  uint32_t timeAfter;
  uint64_t benchmarkingTrainingTime = 0;
  uint64_t benchmarkingPredictionTime = 0;
  Serial.println("Bechmarking start");
  for(uint32_t i=0;i<BENCHMARK_AVERAGING;i++){
    TinyDecisionTreeClassifier<Treetype> clf(MAX_TREE_DEPTH,MIN_SAMPLES_SPLIT);
    timeBefore = micros();
    clf.fit(X_train_ptr,Y_train_ptr,NUMBER_OF_TRAIN_SAMPLES,NUMBER_OF_FEATURES);
    // Serial.println("Tree before pruning: ");
    // clf.plot();
    while(clf.prune(X_train_ptr,Y_train_ptr,NUMBER_OF_TRAIN_SAMPLES));
    timeAfter = micros();
    benchmarkingTrainingTime+=(timeAfter-timeBefore);
    Treetype rslt;
    for(uint32_t j=0;j<NUMBER_OF_TRAIN_SAMPLES;j++){
        timeBefore = micros();
        rslt = clf.predict(X_train_ptr[j]);
        timeAfter = micros();
        benchmarkingPredictionTime=(timeAfter-timeBefore);
    }

    clf.plot();
    Serial.print("Score:");
    Serial.println(clf.score(X_test_ptr,Y_test_ptr,NUMBER_OF_TEST_SAMPLES));
  }
  Serial.print("Training time: ");
  Serial.print((uint32_t)benchmarkingTrainingTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("Prediction time: ");
  Serial.print((uint32_t)benchmarkingPredictionTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("After averaging ");
  Serial.print(BENCHMARK_AVERAGING);
  Serial.println(" sessions.");
}

void readTrain(fs::FS &fs, const char * path){
  Serial.printf("Reading train file: %s\n", path);

  File file = fs.open(path);
  if(!file){
    Serial.println("Failed to open file for reading");
    return;
  }

  Serial.print("Read from file: ");
  uint32_t i=0;
    while (file.available()) {                // Check if more data is available to read
      String line = file.readStringUntil('\n'); // Read a line from the file
      if(!line.isEmpty()){
        if(sscanf(line.c_str(),"%f,%f,%f,%f,%f\n",&X_train_dat[i][0],&X_train_dat[i][1],&X_train_dat[i][2],&X_train_dat[i][3],&Y_train_dat[i][0])){
        Serial.print(X_train_dat[i][0]);
        Serial.print(",");
        Serial.print(X_train_dat[i][1]);
        Serial.print(",");
        Serial.print(X_train_dat[i][2]);
        Serial.print(",");
        Serial.print(X_train_dat[i][3]);
        Serial.print("->");
        Serial.println(Y_train_dat[i][0]);
        i++;
        }
      }
  }
  file.close();
}

void readTest(fs::FS &fs, const char * path){
  Serial.printf("Reading test file: %s\n", path);

  File file = fs.open(path);
  if(!file){
    Serial.println("Failed to open file for reading");
    return;
  }

  Serial.print("Read from file: ");
  uint32_t i=0;
  int32_t idx;
  while (file.available()) {                // Check if more data is available to read
      String line = file.readStringUntil('\n'); // Read a line from the file
      if(!line.isEmpty()){
        if(sscanf(line.c_str(),"%ld,%f,%f,%f,%f,%f\n",&idx,&X_test_dat[i][0],&X_test_dat[i][1],&X_test_dat[i][2],&X_test_dat[i][3],&Y_test_dat[i][0])){
        Serial.print(X_train_dat[i][0]);
        Serial.print(",");
        Serial.print(X_train_dat[i][1]);
        Serial.print(",");
        Serial.print(X_train_dat[i][2]);
        Serial.print(",");
        Serial.print(X_train_dat[i][3]);
        Serial.print("->");
        Serial.println(Y_train_dat[i][0]);
        i++;
        }
      }
  }
  file.close();
}

void setup() {
  Serial.begin(115200);
  delay(1000);
  uint32_t i;
  X_train_ptr = new Treetype *[NUMBER_OF_TRAIN_SAMPLES];
  Y_train_ptr = new Treetype *[NUMBER_OF_TRAIN_SAMPLES];
  for(i=0;i<NUMBER_OF_TRAIN_SAMPLES;i++){
      X_train_ptr[i] = X_train_dat[i];
      Y_train_ptr[i] = Y_train_dat[i];
  }

  X_test_ptr = new Treetype *[NUMBER_OF_TEST_SAMPLES];
  Y_test_ptr = new Treetype *[NUMBER_OF_TEST_SAMPLES];
  for(i=0;i<NUMBER_OF_TEST_SAMPLES;i++){
      X_test_ptr[i] = X_test_dat[i];
      Y_test_ptr[i] = Y_test_dat[i];
  }

  SPI.begin(18,19,23,5);
  if(!SD.begin(5)){
    Serial.println("Card Mount Failed");
    return;
  }

  readTrain(SD,"/iris_train_ok.csv");
  readTest(SD,"/iris_test_ok.csv");
  benchmark();
}

void loop() {
  delay(100);
}
