#include <Arduino.h>
// #define DTR_DEBUG_
#include "TinyDecisionTreeClassifier.h"

#define NUMBER_OF_FEATURES  5
#define NUMBER_OF_SAMPLES   7
#define MAX_TREE_DEPTH      3
typedef float Treetype;
Treetype X_const[NUMBER_OF_SAMPLES][NUMBER_OF_FEATURES] =
{ {1,2,3,1,1},
  {1,2,1,1,1},
  {1,1,2,3,1},
  {1,2,3,1,2},
  {3,2,1,1,1},
  {1,2,3,1,2},
  {3,2,3,1,2},
  };

Treetype Y_const[NUMBER_OF_SAMPLES][1] = {
  {1},
  {0},
  {1},
  {0},
  {1},
  {1},
  {0},
  };

Treetype **X;
Treetype **Y;

void fillBuffersWithRandom(void){
  for(uint32_t i=0;i<NUMBER_OF_SAMPLES;i++){
    for(uint32_t j=0;j<NUMBER_OF_FEATURES;j++){
        X_const[i][j]=(Treetype)random(-128,127);
      }
      Y_const[i][0] = (Treetype)random(0,2);
  }
}

#define BENCHMARK_AVERAGING   1
void benchmark(void){
  uint32_t timeBefore;
  uint32_t timeAfter;
  uint64_t benchmarkingTrainingTime = 0;
  uint64_t benchmarkingPredictionTime = 0;
  Serial.println("Bechmarking start");
  for(uint32_t i=0;i<BENCHMARK_AVERAGING;i++){
    TinyDecisionTreeClassifier<Treetype> clf(MAX_TREE_DEPTH,2);
    timeBefore = micros();
    clf.fit(X,Y,NUMBER_OF_SAMPLES,NUMBER_OF_FEATURES);
    Serial.println("Tree before pruning: ");
    clf.plot();
    while(clf.prune(X,Y,NUMBER_OF_SAMPLES));
    timeAfter = micros();
    benchmarkingTrainingTime+=(timeAfter-timeBefore);
    Treetype rslt;
    for(uint32_t j=0;j<NUMBER_OF_SAMPLES;j++){
        timeBefore = micros();
        rslt = clf.predict(X[j]);
        timeAfter = micros();
        benchmarkingPredictionTime=(timeAfter-timeBefore);
    }
  clf.plot();
  }
  Serial.print("Training time: ");
  Serial.print((uint32_t)benchmarkingTrainingTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("Prediction time: ");
  Serial.print((uint32_t)benchmarkingPredictionTime/BENCHMARK_AVERAGING);
  Serial.println(" us");
  Serial.print("After averaging ");
  Serial.print(BENCHMARK_AVERAGING);
  Serial.println(" sessions.");
}

void saveLoadTest(void){
    uint32_t memBefore;
    uint32_t memAfter;

    for(uint32_t k=0;k<100;k++){
      Serial.println("Start");
      fillBuffersWithRandom();
      uint32_t outputdatasize = 100;
      char* outputdata = (char*) malloc(outputdatasize);
      TinyDecisionTreeClassifier<Treetype> clf1(MAX_TREE_DEPTH,2);
      clf1.fit(X,Y,NUMBER_OF_SAMPLES,NUMBER_OF_FEATURES);
      clf1.plot();
      while(clf1.prune(X,Y,NUMBER_OF_SAMPLES));
      clf1.plot();
      clf1.save(&outputdata,&outputdatasize);
      Serial.println("Tree saved, load start");
      Serial.println(outputdata);

      TinyDecisionTreeClassifier<Treetype> clf2(MAX_TREE_DEPTH,2);
      clf2.load(outputdata);
      clf2.plot();
      free(outputdata);
      Serial.println("Finished");

    }
   
}


void setup() {
  Serial.begin(115200); 
  X = new Treetype *[NUMBER_OF_SAMPLES];
  Y = new Treetype *[NUMBER_OF_SAMPLES];
  for(uint32_t i=0;i<NUMBER_OF_SAMPLES;i++){
      X[i] = X_const[i];
      Y[i] = Y_const[i];
  }
  saveLoadTest();
  // clf.plot();
}

void loop() {
  // Serial.println("Begin!");
  // digitalWrite(13,HIGH);
  // delay(100);
  // digitalWrite(13,LOW);
  delay(100);
}
