#include <Arduino.h>
// #define DTR_DEBUG_
#include "TinyDecisionTreeClassifier.h"

#define NUMBER_OF_FEATURES  5
#define NUMBER_OF_SAMPLES   20
#define MAX_TREE_DEPTH      3
typedef float Treetype;
Treetype X_const[NUMBER_OF_SAMPLES][NUMBER_OF_FEATURES];
Treetype Y_const[NUMBER_OF_SAMPLES][1];

Treetype **X;
Treetype **Y;

#define BENCHMARK_AVERAGING   10
void fillBuffersWithRandom(void){
  for(uint32_t i=0;i<NUMBER_OF_SAMPLES;i++){
    for(uint32_t j=0;j<NUMBER_OF_FEATURES;j++){
        X_const[i][j]=(Treetype)random(-128,127);
      }
      Y_const[i][0] = (Treetype)random(0,2);
  }
}

#define MEMORY_LEAKS_TEST_REPEAT   10
void memory_leaks_test(void){
  uint32_t timeBefore;
  uint32_t timeAfter;
  uint32_t heapBefore;
  uint32_t heapAfter;
  uint64_t benchmarkingTrainingTime = 0;
  uint64_t benchmarkingPredictionTime = 0;
  Serial.println("Memory leaks testing start");

  heapBefore = ESP.getFreeHeap();
  for(uint32_t i=0;i<MEMORY_LEAKS_TEST_REPEAT;i++){
    TinyDecisionTreeClassifier<Treetype> clf(MAX_TREE_DEPTH,2);
    fillBuffersWithRandom();
    timeBefore = micros();
    clf.fit(X,Y,NUMBER_OF_SAMPLES,NUMBER_OF_FEATURES);
    timeAfter = micros();
    fillBuffersWithRandom();
    while(clf.prune(X,Y,NUMBER_OF_SAMPLES));
    benchmarkingTrainingTime+=(timeAfter-timeBefore);
    Treetype rslt;
    for(uint32_t j=0;j<NUMBER_OF_SAMPLES;j++){
        timeBefore = micros();
        rslt = clf.predict(X[j]);
        timeAfter = micros();
        benchmarkingPredictionTime=(timeAfter-timeBefore);
    }
  }
  heapAfter = ESP.getFreeHeap();
  if(heapBefore==heapAfter){
    Serial.println("No Memory leaks detected!");
  }else{
    Serial.println("WARNING, memory leak detected!");
  }
}



void setup() {
  Serial.begin(115200); 
  X = new Treetype *[NUMBER_OF_SAMPLES];
  Y = new Treetype *[NUMBER_OF_SAMPLES];
  for(uint32_t i=0;i<NUMBER_OF_SAMPLES;i++){
      X[i] = X_const[i];
      Y[i] = Y_const[i];
  }
  memory_leaks_test();
}

void loop() {
  // Serial.println("Begin!");
  // digitalWrite(13,HIGH);
  // delay(100);
  // digitalWrite(13,LOW);
  delay(100);
}
